import datetime
import json
from pathlib import Path


class ExtendedJsonEncoder(json.JSONEncoder):
    """
    Additional opiniated support for more basic object types.
    """
    def default(self, obj):
        # Support for pathlib.Path to a string
        if isinstance(obj, Path):
            return str(obj)
        # Support for set to a list
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def dump_datasets(basedir, source_name, source, result, resolved_payload):
    """
    A helper function to dump source and resolving result lists in JSON files.
    """
    apps_filename = result.replace("_result", "_apps")

    source_destination = basedir / source_name
    result_destination = basedir / result
    apps_destination = basedir / apps_filename

    print("source_destination:", source_destination)
    print("result_destination:", result_destination)
    print("apps_destination:", apps_destination)

    # Write the collection source
    # print("SRC-"*30)
    # print()
    # print(source)
    source_destination.write_text(source)

    # Write the resolved collection as payload items
    result_content = json.dumps(resolved_payload, indent=4)
    # print("PAYLOAD-"*20)
    # print()
    # print(result_content)
    result_destination.write_text(result_content)

    # Write the resolved collection as just application names
    apps = json.dumps([item.get("name") for item in resolved_payload], indent=4)
    # print("APPS-"*30)
    # print()
    # print(apps)
    apps_destination.write_text(apps)

    return source_destination, result_destination, apps_destination



import sys
import inspect

from ..app_storage import AppStore
from ..exceptions import ComposerError
from ..importer import import_module
from ..logger import LoggerBase
from ..manifest import Manifest


class ComposerBase(LoggerBase):
    """
    Composer base implements everything about application module discovering and
    manifest loading.

    Arguments:
        manifest (string or pathlib.Path or dict or Manifest): The Manifest source to
            load. It can be either:

            * A Manifest object, it will just returned as it without any validation,
                you are responsible of its correctness;
            * A string for the file path to load in JSON or TOML format;
            * A Path object to the file to load in JSON or TOML format;
            * A Dictionnary which respect the manifest structure;

            Source file format are guessed from their file extension such as JSON
            for ``.json`` or TOML for ``.toml``.

    Keyword Arguments:
        syspaths (list): A list of path to insert in sys.path, this is required if
            ``repository`` module is not available directly in the current working
            directory. You are responsible to no set twice paths, erroneous paths, etc..
            There is no validation for the paths you give.

    Attributes:
        _MODULE_PYTHONPATH (string): A template string to build the full Python path
            of founded class. It expected two variable ``parent`` and ``name``,
            respectively the module path and the class name.
    """
    _MODULE_PYTHONPATH = "{parent}.{name}"

    def __init__(self, manifest, **kwargs):
        super().__init__()

        self.manifest = self.get_manifest(manifest)
        self.set_syspaths(self.manifest.syspaths or [])
        self.apps = self.resolve_collection(manifest)

    def get_manifest(self, manifest):
        """
        Return loaded manifest object.

        Arguments:
            manifest (string or pathlib.Path or dict or Manifest): The Manifest source
                to load.

        Returns:
            Manifest: The manifest object loaded from given source.
        """
        return Manifest.load(manifest)

    def set_syspaths(self, paths):
        """
        Append each item path to ``sys.path``.

        This won't never append a same path twice.

        Arguments:
            paths (list): A list of path to append.
        """
        for path in paths:
            if path not in sys.path:
                sys.path.append(path)

    def get_module_path(self, name):
        """
        Return a Python path for a module name.

        Arguments:
            name (string): Module name.

        Returns:
            string: Module name prefixed with repository path if it is not empty else
            returns just the module name.
        """
        if self.manifest.repository:
            return self._MODULE_PYTHONPATH.format(
                parent=self.manifest.repository,
                name=name,
            )

        return name

    def find_app_module(self, name):
        """
        Find a module (by its pythonpath) from application.

        Arguments:
            name (string): Module pythonpath.

        Returns:
            object: Module object if found else None.
        """
        try:
            module = import_module(name)
        except ModuleNotFoundError:
            msg = "{klass} is unable to find module: {path}".format(
                klass=self.__class__.__name__,
                path=name,
            )
            self.log.warning(msg)
            return None
        else:
            return module

    def _is_elligible_class(self, obj):
        """
        Find if given object is an enabled class for composition.

        Criterias for eligibility are in order:

        * Object is a class;
        * Class is not named ``EnabledApplicationMarker``;
        * Class got attribute ``_ENABLED_COMPOSABLE_APPLICATION`` which value is not
          ``None``;

        Arguments:
            obj (object): Object to check for eligibility.

        Returns:
            boolean: ``True`` if object is eligibile to criterias else ``False``.
        """
        if (
            inspect.isclass(obj) and
            getattr(obj, "__name__", None) != "EnabledApplicationMarker" and
            getattr(obj, "_ENABLED_COMPOSABLE_APPLICATION", None) is not None
        ):
            return True

        return False

    def _get_elligible_module_classes(self, path, module):
        """
        Get all elligible classes from a module.

        Arguments:
            path (string): The Python path to a module used for reporting and logging
                messages.
            module (object): The module object where to find elligible classes.

        Returns:
            list: List of elligible classes objects.
        """
        enabled = []

        if not hasattr(module, "__dict__"):
            msg = "Module object from '{}' must have a '__dict__' attribute."
            raise ComposerError(msg.format(path))

        for object_name in module.__dict__.keys():
            if not object_name.startswith("_"):
                obj = getattr(module, object_name)

                if self._is_elligible_class(obj):
                    msg = "{klass} found enabled Class at: {path}.{object_name}".format(
                        klass=self.__class__.__name__,
                        path=path,
                        object_name=object_name,
                    )
                    self.log.debug(msg)

                    if obj not in enabled:
                        enabled.append(obj)

        return enabled

    def _scan_app_module(self, name):
        """

        Arguments:
            name (string):

        Returns:
            dict:
        """
        path = self.get_module_path(name)

        # Try to find application module and get its possible parameter variables
        module = self.find_app_module(path)
        if module:
            msg = "{klass} found application at: {path}".format(
                klass=self.__class__.__name__,
                path=path,
            )
            self.log.debug(msg)

            payload = {"name": name}

            if hasattr(module, "DEPENDENCIES"):
                payload["dependencies"] = getattr(module, "DEPENDENCIES")

            if hasattr(module, "PUSH_END"):
                payload["push_end"] = getattr(module, "PUSH_END")

            return payload

        return None

    def resolve_collection(self, manifest):
        """
        Resolve collection with AppStore.

        Arguments:
            manifest (Manifest): The Manifest object.

        Returns:
            list: List of AppNode.
        """
        collection = []

        for name in self.manifest.collection:
            payload = self._scan_app_module(name)
            # Ignore unfound application
            if payload:
                collection.append(payload)

        store = AppStore(
            default_app=self.manifest.default_store_app,
            no_ordering=self.manifest.no_ordering,
        )
        return store.resolve(collection)



"""
=====================
Django parts composer
=====================

"""
from .classes import ClassComposer


class ComposeDjangoSettings(ClassComposer):
    """
    Composer for enabled application settings classes for a Django project.
    """
    _MODULE_PYTHONPATH = "{parent}.{name}.settings"


class ComposeDjangoUrls(ClassComposer):
    """
    Composer for enabled application urls classes for a Django project.
    """
    _MODULE_PYTHONPATH = "{parent}.{name}.urls"




class EnabledApplicationMarker:
    """
    Empty class to mark a settings class mixin as enabled to load for composition.

    This class won't never ever introduce anything else than marker attribute
    ``_ENABLED_COMPOSABLE_APPLICATION`` since the class may be herited by any kind of
    class we don't want to pollute.

    Only purpose of attribute ``_ENABLED_COMPOSABLE_APPLICATION`` is to mark it for
    inspection from composer.
    """
    _ENABLED_COMPOSABLE_APPLICATION = True



from .base import ComposerBase
from .classes import ClassComposer
from .text import TextContentComposer
from .django import ComposeDjangoSettings, ComposeDjangoUrls
from .purge import PurgeApplications


__all__ = [
    "ComposerBase",
    "ClassComposer",
    "TextContentComposer",
    "ComposeDjangoSettings",
    "ComposeDjangoUrls",
    "PurgeApplications",
]



import datetime

from pathlib import Path

from .base import ComposerBase


class TextContentComposer(ComposerBase):
    """
    Text content composer assemble all text content files from enabled Applications.
    """
    def get_template(self, template=None):
        """
        Get the base content text where to append applications requirements.

        Keyword Arguments:
            template (string or pathlib.Path): Path to file of base content to start
                output.

        Returns:
            string: Base content.
        """
        if template:
            template = Path(template)
            return template.read_text()

        return ""

    def export(self):
        """
        Combinate all application content files into a single content string.

        Returns:
            string: Combinated content files.
        """
        output = ""

        if self.manifest.requirements.introduction:
            output += self.manifest.requirements.introduction.format(
                creation_date=datetime.datetime.now().isoformat(timespec="seconds"),
            )

        output += self.get_template(self.manifest.requirements.template)

        for node in self.apps:
            # Try to find application module
            module_path = self.get_module_path(node.name)
            module = self.find_app_module(module_path)

            if module and getattr(module, "__file__", None):
                # Resolve expected text content file path inside module
                source_path = (
                    Path(module.__file__).parents[0].resolve() /
                    self.manifest.requirements.source_filename
                )
                # Try to find file from application to append its content to the output
                if source_path.exists():
                    msg = "{klass} found content file at: {path}".format(
                        klass=self.__class__.__name__,
                        path=source_path,
                    )
                    self.log.debug(msg)

                    content = source_path.read_text()
                    if content.strip():
                        if self.manifest.requirements.application_divider:
                            output += self.manifest.requirements.application_divider

                        if self.manifest.requirements.application_label:
                            label = self.manifest.requirements.application_label
                            output += label.format(name=node.name)

                        output += content

                else:
                    msg = "{klass} is unable to find content file from: {path}".format(
                        klass=self.__class__.__name__,
                        path=source_path,
                    )
                    self.log.debug(msg)

        return output

    def dump(self, destination):
        """
        Write export payload to a dump file.

        Arguments:
            destination (pathlib.Path): Path object for the dump file destination.

        Returns:
            pathlib.Path: The Path object where the file has been writed.
        """
        output = self.export()
        destination.write_text(output)

        return destination




def debug_invoke(result, caplog=None):
    """
    Output important informations on Click invoke result and optionally caplog.

    It invoke command has raised an exception, raise it up.
    """
    print("=> result.output <=")
    print(result.output)
    print()

    if caplog:
        print("=> caplog.record_tuples <=")
        print(caplog.record_tuples)
        print()

    print("=> result.exception <=")
    print(result.exception)
    if result.exception is not None:
        raise result.exception
